from typing import Union

from .combinatorics import permutation_with_repeat
from .string import replacing
from .exceptions import ImpossibleValue


def print_true_table(variables: str, expretion: str, value: Union[int, bool, 'all'] = 'all') -> None:
    """
    Вывод таблицы истинности лог.функции expretion от переменных variables.

    Если value='all', выводятся все строки таблицы;

    Если value=1 или True, выводятся строки таблицы, где функция истинна;

    Если value=0 или False, выводятся строки таблицы, где функция ложна.

    В качестве лог.операций можно использовать обычные операции Python
    или такие эквиваленты:
    {"&": " and ", "|": " or ", "~": " not ", "->": "<="}
    """
    if not isinstance(variables, str):
        _type = str(type(variables))
        raise TypeError("'variables' должен иметь тип str, а передан тип {}!".format(_type.split()[1][1:-2]))

    if not isinstance(expretion, str):
        _type = str(type(expretion))
        raise TypeError("'expretion' должен иметь тип str, а передан тип {}!".format(_type.split()[1][1:-2]))

    if not any(isinstance(value, t) for t in (int, bool, str)):
        _type = str(type(value))
        raise TypeError("'value' должен иметь тип int, bool или str, а передан тип {}!".format(_type.split()[1][1:-2]))

    if isinstance(value, str) and value != "all":
        raise ImpossibleValue("Неизвестное {}. Возможно, Вы имели в виду \"all\"!".format(value))

    operations = {"&": " and ", "|": " or ", "~": " not ", "->": "<="}

    for new, old in operations.items():
        expretion = expretion.replace(new, old)

    print(variables, 'F')

    for row in permutation_with_repeat('01', len(variables)):
        copy_exp = expretion

        for old, new in zip(variables, row):
            copy_exp = replacing(copy_exp, old, new, 'целиком')

        if (result := int(eval(copy_exp))) == value:
            print(*row, ' ', value, sep='')

        if value == 'all':
            print(*row, ' ', result, sep='')



from .decorators import *
from .mathematics import *
from .constants import *
from .algebra_logics import *
from .combinatorics import *
from .system_count import *
from .string import *
from .lists_and_other import *

__author__ = "Ilya484"



from typing import Union
from .exceptions import InvalidBase

__all__ = ["to_base"]


def __validate(num, ob, nb) -> None:
    if not any(isinstance(num, t) for t in (int, str)):
        _type_num = str(type(num))
        raise TypeError("'number' должен иметь тип int или str, а передан тип {}!".format(_type_num.split()[1][1:-2]))
    if not isinstance(ob, int):
        _type_old_base = str(type(ob))
        raise TypeError("'old_base' должен иметь тип int, а передан тип {}!".format(_type_old_base.split()[1][1:-2]))
    if not isinstance(nb, int):
        _type_new_base = str(type(nb))
        raise TypeError("'new_base' должен иметь тип int, а передан тип {}!".format(_type_new_base.split()[1][1:-2]))


def to_base(number: Union[int, str], old_base: int = 10, new_base: int = 10) -> Union[int, str]:
    """
    Переводит число number с основанием old_base в число с основанием new_base.
    """
    __validate(number, old_base, new_base)
    if old_base == new_base:
        return number
    elif not (2 <= old_base <= 36) or not (2 <= new_base <= 36):
        raise InvalidBase("Невозможно перевести, т.к., old_base и new_base должны быть от 2 до 36")
    elif new_base == 10:
        return int(str(number), old_base)
    elif old_base == 10:
        mods = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'
        r = ''
        while number:
            number, y = divmod(number, new_base)
            r = mods[y] + r
        return r
    else:
        return to_base(to_base(number, old_base), 10, new_base)




from typing import Union

from .system_count import to_base
from .exceptions import ImpossibleValue

__all__ = ["permutation_with_repeat", "permutations"]


def permutation_with_repeat(seq: Union[list, tuple, str], repeat: int = 1):
    """
    Возвращает перестановки элкементов итерируемого
    обьекта seq с repeat повторениями.
    """
    if not any(isinstance(seq, t) for t in (list, tuple, str)):
        _type = str(type(seq))
        raise TypeError("'seq' должен иметь тип list, tuple или str, а передан тип {}!".format(_type.split()[1][1:-2]))

    if not isinstance(repeat, int):
        _type = str(type(repeat))
        raise TypeError("'repeat' должен иметь тип int, а передан тип {}!".format(_type.split()[1][1:-2]))

    if repeat <= 0:
        raise ImpossibleValue("'repeat' <= 0, но число повторений должно быть положительным!")

    def _gen_mask():
        for v in range(len(seq) ** repeat):
            n = to_base(v, new_base=len(seq))
            yield '0' * (repeat - len(n)) + n

    for m in _gen_mask():
        s = [seq[int(i)] for i in m]
        yield tuple(s)


def permutations(seq: Union[list, tuple, str]):
    """
    Возвращает перестановки элкементов итерируемого объекта seq.
    """
    if not any(isinstance(seq, t) for t in (list, tuple, str)):
        _type = str(type(seq))
        raise TypeError("'seq' должен иметь тип list, tuple или str, а передан тип {}!".format(_type.split()[1][1:-2]))

    if len(seq) == 0:
        return []
    if len(seq) == 1:
        return [seq]

    lst = []

    for i in range(len(seq)):
        m = seq[i]
        remLst = seq[:i] + seq[i + 1:]
        for p in permutations(remLst):
            lst.append(m + p)
    return lst



from typing import Union
from .exceptions import InvalidMode, ImpossibleValue

__all__ = ["replacing", "index_n", "is_number"]


def replacing(string: str, substring: str, new_string: str, mode: str = 'обычно',
              cnt: Union[int, str] = 'all') -> str:
    """
    Возвращает строку string с заменённой подстрокой
    substring на  подстроку new_string в количестве
    cnt.

    Режим "обычно":
                    замена стандартным replace

    Режим "целиком":
                    замена подстроки substring если она не
                    является частью большей подстроки.
    """
    if not isinstance(string, str):
        _type = str(type(string))
        raise TypeError("'string' должен иметь тип str, а передан тип {}!".format(_type.split()[1][1:-2]))

    if not isinstance(substring, str):
        _type = str(type(substring))
        raise TypeError("'substring' должен иметь тип str, а передан тип {}!".format(_type.split()[1][1:-2]))

    if not isinstance(new_string, str):
        _type = str(type(new_string))
        raise TypeError("'new_string' должен иметь тип str, а передан тип {}!".format(_type.split()[1][1:-2]))

    if not isinstance(mode, str):
        _type = str(type(mode))
        raise TypeError("'mode' должен иметь тип str, а передан тип {}!".format(_type.split()[1][1:-2]))

    if mode not in ("обычно", "целиком"):
        raise InvalidMode("Неизвестный режим {}. Режим должен быть \"обычно\" или \"целиком\"!".format(mode))

    if cnt != "all" and isinstance(cnt, str):
        raise ImpossibleValue("'cnt' имеет тип str, но отличное от \"all\"!")

    if isinstance(cnt, int) and cnt < 0:
        raise ImpossibleValue("'cnt' < 0, но количество не может быть отрительным!")

    if cnt == 0:
        return string

    if mode == 'обычно':
        if cnt == 'all':
            return string.replace(substring, new_string)
        return string.replace(substring, new_string, cnt)

    if mode == 'целиком':
        result = ''
        if string[:len(substring)] == substring and not string[len(substring)].isalpha():
            string = string[len(substring):]
            result = new_string
        else:
            result += string[0]
            string = string[1:]

        while len(string):
            if len(string) < len(substring):
                result += string
                break
            else:
                try:
                    next_char = string[len(substring)]
                    if string[:len(substring)] == substring and not result[-1].isalpha() and not next_char.isalpha():
                        result += new_string
                        string = string[len(substring):]
                    else:
                        result += string[0]
                        string = string[1:]
                except IndexError:
                    if string == substring:
                        result += new_string
                        break
                    else:
                        result += string
                        break
        return result


def index_n(string: str, substring: str, n: int = 1) -> int:
    """
    Возвращает индекс n-го вхождения СЛЕВА подстроки
    substring в строку string. Если такого вхождения
    нет, возвращается -1000.
    """
    if not isinstance(string, str):
        _type = str(type(string))
        raise TypeError("'string' должен иметь тип str, а передан тип {}!".format(_type.split()[1][1:-2]))

    if not isinstance(substring, str):
        _type = str(type(substring))
        raise TypeError("'substring' должен иметь тип str, а передан тип {}!".format(_type.split()[1][1:-2]))

    if not isinstance(n, int):
        _type = str(type(n))
        raise TypeError("'n' должен иметь тип int, а передан тип {}!".format(_type.split()[1][1:-2]))

    if n <= 0:
        raise ImpossibleValue("'n' <= 0, но номер вхождения должен быть положительным!")

    index, cnt = -1000, 0
    for i in range(len(string) - len(substring) + 1):
        if string[i:i + len(substring):] == substring:
            cnt += 1
            if cnt == n:
                index = i
                break
    return index


def is_number(n: str) -> bool:
    """
    Проверяет является ли строка n числом.
    Если да возвращается True, иначе - False.
    """
    if not isinstance(n, str):
        _type = str(type(n))
        raise TypeError("'n' должен иметь тип str, а передан тип {}!".format(_type.split()[1][1:-2]))

    return all("0" <= sym <= "9" for sym in n)



from typing import Union


def prod(seq: Union[list, tuple, set]) -> Union[int, float]:
    """
    Возввращает произведение элементов итерируемого объекта seq.
    """
    if not any(isinstance(seq, t) for t in (list, tuple, set)):
        _type = str(type(seq))
        raise TypeError("'seq' должен иметь тип list, tuple или set, а передан тип {}!".format(_type.split()[1][1:-2]))

    p = 1

    for el in seq:
        if not any(isinstance(el, t) for t in (int, float)):
            _type = str(type(el))
            raise TypeError(
                            "'seq' содержит не поддерживаемый тип {}!".format(_type.split()[1][1:-2]))
        if el == 0:
            return 0
        p *= el

    return p



from typing import Callable


def cache(f: Callable) -> Callable:
    """
    Кэширование функции (декоратор).
    """
    ch = {}

    def wrapper(*args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in ch:
            ch[key] = f(*args, **kwargs)
        return ch[key]

    wrapper.__name__ = f.__name__
    wrapper.__doc__ = f.__doc__
    return wrapper



class InvalidBase(ValueError):
    __module__ = Exception.__module__


class InvalidMode(ValueError):
    __module__ = Exception.__module__


class ImpossibleValue(ValueError):
    __module__ = Exception.__module__



import setuptools

long_description = """# Документация
## Библиотека работает в Python 3.6+ 
### Немного терминологии
**Модуль** - это файл с расширением `.py`, находящийся в библиотеке.

Например, модуль `constants` - это файл `constants.py`.

---
### Модуль decorators
#### 1. Декоратор cache
Для чего: *мемоизация работы функции*.

Пример использования:
```python
from infEGE import cache

@cache
def fib(n):
    if n in {0,1}:
        return 1
    return fib(n - 1) + fib(n - 2)

fib(50)
```
Если убрать `@cache`, то вычисляться будет очень долго.

---
### Модуль constants
Для чего: *константы для использования в алгоритмах*.

Содержимое:
```python
E = 2.718281828459045
PI = 3.141592653589793
maxValue = float('inf')
minValue = float('-inf')
```
---

### Модуль algebra_logic
#### 1. Функция print_true_table
Синтаксис: **print_true_table(variables: str, expretion: str, value: Union[int, bool, 'all'] = 'all') -> None**

Для чего: 
    
    Вывод таблицы истинности лог.функции expretion от переменных variables.

    Если value='all', выводятся все строки таблицы;

    Если value=1 или True, выводятся строки таблицы, где функция истинна;

    Если value=0 или False, выводятся строки таблицы, где функция ложна.

    В качестве лог.операций можно использовать обычные операции Python
    или такие эквиваленты:
    {"&": " and ", "|": " or ", "~": " not ", "->": "<="}.
Пример использования #1:
```python
from infEGE import print_true_table

print_true_table("xy", "x->y", 1)
```
Вывод:
```
xy F
00 1
01 1
11 1
```
**Внимание: в Python приоритет <= выше, чем and, or, not! Ставьте скобки!** 

Пример использования #2:
```python
from infEGE import print_true_table

print_true_table("xy", "x&(y|x)|y", 0)
```
Вывод:
```
xy F
00 0
```

Пример использования #3:
```python
from infEGE import print_true_table

print_true_table("xzy", "x or z and (y or not x)")
```
Вывод:
```
xzy F
000 0
001 0
010 1
011 1
100 1
101 1
110 1
111 1
```
---
### Модуль combinatorics
#### 1. Функция permutation_with_repeat
Синтаксис: **permutation_with_repeat(seq: Union[list, tuple, str], repeat: int = 1)**

Для чего: 

    Возвращает перестановки элкементов итерируемого
    обьекта seq с repeat повторениями.
Пример использования:

```python
from infEGE import permutation_with_repeat

for el in permutation_with_repeat('123', 2):
    print(el)
```
Вывод:
```
('1', '1')
('1', '2')
('1', '3')
('2', '1')
('2', '2')
('2', '3')
('3', '1')
('3', '2')
('3', '3')
```
#### 2. Функция permutations
Синтаксис: **permutations(seq: Union[list, tuple, str]):**

Для чего: 

    Возвращает перестановки элкементов итерируемого объекта seq.
Пример использования:
```python
from infEGE import permutations

for el in permutations('abc'):
    print(el)
```
Вывод:
```
abc
acb
bac
bca
cab
cba
```
---
### Модуль lists_and_other
#### 1. Функция prod
Синтаксис: **prod(seq: Union[list, tuple, set]) -> Union[int, float]**

Для чего: 

    Возввращает произведение элементов итерируемого объекта seq.
Пример использования:
```python
from infEGE import prod

print(prod([5, 8, 6, 100, 54]))
```
Вывод:
```
1296000
```
---
### Модуль string
#### 1. Функция replacing
Синтаксис: **replacing(string: str, substring: str, new_string: str, mode: str = 'обычно', cnt: Union[int, str] = 'all') -> str**

Для чего: 

    Возвращает строку string с заменённой подстрокой
    substring на  подстроку new_string в количестве cnt.

    Режим "обычно":
                    замена стандартным replace

    Режим "целиком":
                    замена подстроки substring если она не
                    является частью большей подстроки.
Пример использования #1:
```python
from infEGE import replacing

print(replacing("Питон плохой тон", "тон", "нот"))
```
Вывод:
```
Пинот плохой нот
```

Пример использования #2:
```python
from infEGE import replacing

print(replacing("Питон плохой тон", "тон", "нот", cnt=1))
```
Вывод:
```
Пинот плохой тон
```

Пример использования #3:
```python
from infEGE import replacing

print(replacing("Питон плохой тон", "тон", "нот", "целиком"))
```
Вывод:
```
Питон плохой нот
```

#### 2. Функция index_n
Синтаксис: **index_n(string: str, substring: str, n: int = 1) -> int**

Для чего: 

    Возвращает индекс n-го вхождения СЛЕВА подстроки
    substring в строку string. Если такого вхождения
    нет, возвращается -1000.
Пример использования #1:
```python
from infEGE import index_n

print(index_n("01230123", "1"))
```
Вывод:
```
1
```

Пример использования #2:
```python
from infEGE import index_n

print(index_n("01230123", "1", 2))
```
Вывод:
```
5
```

Пример использования #3:
```python
from infEGE import index_n

print(index_n("01230123", "1", 3))
```
Вывод:
```
-1000
```
#### 3. Функция is_number
Синтаксис: **is_number(n: str) -> bool**

Для чего: 

    Проверяет является ли строка n числом.
    Если да возвращается True, иначе - False.
Пример использования #1:
```python
from infEGE import is_number

print(is_number("23"))
```
Вывод:
```
True
```

Пример использования #2:
```python
from infEGE import is_number

print(is_number("2n3"))
```
Вывод:
```
False
```
---
### Модуль system_count
#### 1. Функция to_base
Синтаксис: **to_base(number: Union[int, str], old_base: int = 10, new_base: int = 10) -> Union[int, str]**

Для чего: 

    Переводит число number с основанием old_base в число
    с основанием new_base.
Пример использования #1:
```python
from infEGE import to_base

print(to_base(5, new_base=2))
```
Вывод:
```
101
```
Пример использования #2:
```python
from infEGE import to_base

print(to_base(15, new_base=16))
```
Вывод:
```
F
```
Пример использования #3:
```python
from infEGE import to_base

print(to_base("FA32", old_base=17, new_base=10))
```
Вывод:
```
76638
```
Пример использования #4:
```python
from infEGE import to_base

print(to_base("FA32", old_base=17, new_base=6))
```
Вывод:
```
1350450
```
---
### Модуль mathematics
#### 1. Функция is_prime
Синтаксис: **is_prime(n: int) -> bool**

Для чего: 

    Если n - простое, то возващается True, иначе - False.
Пример использования #1:
```python
from infEGE import is_prime

print(is_prime(5))
```
Вывод:
```
True
```
Пример использования #2:
```python
from infEGE import is_prime

print(is_prime(25))
```
Вывод:
```
False
```
Пример использования #3:
```python
from infEGE import is_prime

print(is_prime(1))
```
Вывод:
```
False
```
---
#### 2. Функция is_even
Синтаксис: **is_even(n: int) -> bool**

Для чего: 

    Если n - чётно, то возващается True, иначе - False.
Пример использования #1:
```python
from infEGE import is_even

print(is_even(12))
```
Вывод:
```
True
```
Пример использования #2:
```python
from infEGE import is_even

print(is_even(25))
```
Вывод:
```
False
```
---
#### 3. Функция is_odd
Синтаксис: **is_odd(n: int) -> bool**

Для чего: 

    Если n - нечётно, то возващается True, иначе - False.
Пример использования #1:
```python
from infEGE import is_odd

print(is_odd(12))
```
Вывод:
```
False
```
Пример использования #2:
```python
from infEGE import is_odd

print(is_odd(25))
```
Вывод:
```
True
```
---
#### 4. Функция divided
Синтаксис: **divided(n: int, d: int) -> bool**

Для чего: 

    Если n нацело делится на d, то возвращается True, иначе - False.
Пример использования #1:
```python
from infEGE import divided

print(divided(12, 5))
```
Вывод:
```
False
```
Пример использования #2:
```python
from infEGE import divided

print(divided(121, 11))
```
Вывод:
```
True
```
---
#### 5. Функция not_divisible
Синтаксис: **not_divisible(n: int, d: int) -> bool**

Для чего: 

    Если n не делится нацело на d, то возвращается True, иначе - False.
Пример использования #1:
```python
from infEGE import not_divisible

print(not_divisible(12, 5))
```
Вывод:
```
True
```
Пример использования #2:
```python
from infEGE import not_divisible

print(not_divisible(121, 11))
```
Вывод:
```
False
```
---
#### 6. Функция factorial
Синтаксис: **factorial(n: int) -> int**

Для чего: 

    Возвращает n! (0! = 1)
Пример использования:
```python
from infEGE import factorial

print(factorial(6))
```
Вывод:
```
720
```
---
#### 7. Функция factorize
Синтаксис: **factorize(number: int) -> list**

Для чего: 

    Возвращает разложение числа number на простые множители в list.
Пример использования #1:
```python
from infEGE import factorize

print(factorize(1))
```
Вывод:
```
[]
```
Пример использования #2:
```python
from infEGE import factorize

print(factorize(11))
```
Вывод:
```
[11]
```
Пример использования #3:
```python
from infEGE import factorize

print(factorize(55))
```
Вывод:
```
[5, 11]
```
---
#### 8. Функция divisors
Синтаксис: **divisors(n: int) -> list**

Для чего: 

    Возвращает все натуральные делители числа n на интервале (1; n).
Пример использования #1:
```python
from infEGE import divisors

print(divisors(1))
```
Вывод:
```
[]
```
Пример использования #2:
```python
from infEGE import divisors

print(divisors(720))
```
Вывод:
```
[2, 3, 4, 5, 6, 8, 9, 10, 12, 15, 16, 18, 20, 24, 30, 36, 40, 45, 48, 60, 72, 80, 90, 120, 144, 180, 240, 360]
```
---
#### 9. Функция fib
Синтаксис: **fib(n: int) -> int**

Для чего: 

    Возвращает n-ый член последовательности Фибоначчи. Нумерация с 0.
Пример использования #1:
```python
from infEGE import fib

print(fib(0))
```
Вывод:
```
0
```
Пример использования #2:
```python
from infEGE import fib

print(fib(1))
```
Вывод:
```
1
```
Пример использования #3:
```python
from infEGE import fib

print(fib(2))
```
Вывод:
```
1
```
Пример использования #4:
```python
from infEGE import fib

print(fib(3))
```
Вывод:
```
2
```
Пример использования #5:
```python
from infEGE import fib

print(fib(2001))
```
Вывод:
```
6835702259575806647045396549170580107055408029365524565407553367798082454408054014954534318953113802726603726769523447478238192192714526677939943338306101405105414819705664090901813637296453767095528104868264704914433529355579148731044685634135487735897954629842516947101494253575869699893400976539545740214819819151952085089538422954565146720383752121972115725761141759114990448978941370030912401573418221496592822626
```
**Примечание**: Данный алгоритм работает быстрее рекурсивного! Асимптоматика: O(N)
"""

setuptools.setup(
    name="infEGE",
    version="1.9.7",
    author="Ilya484",
    author_email="ucdo854@kemcdo.ru",
    description="Библиотека для ЕГЭ по информатике",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Ilya484/infEGE",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
)


from math import sqrt

from .exceptions import ImpossibleValue


def divided(n: int, d: int) -> bool:
    """
    Если n нацело делится на d, то возващается True, иначе - False.
    """
    if not isinstance(n, int):
        _type = str(type(n))
        raise TypeError("'n' должен иметь тип int, а передан тип {}!".format(_type.split()[1][1:-2]))

    if not isinstance(d, int):
        _type = str(type(d))
        raise TypeError("'d' должен иметь тип int, а передан тип {}!".format(_type.split()[1][1:-2]))

    return n % d == 0


def not_divisible(n: int, d: int) -> bool:
    """
    Если n не делится нацело на d, то возващается True, иначе - False.
    """
    if not isinstance(n, int):
        _type = str(type(n))
        raise TypeError("'n' должен иметь тип int, а передан тип {}!".format(_type.split()[1][1:-2]))

    if not isinstance(d, int):
        _type = str(type(d))
        raise TypeError("'d' должен иметь тип int, а передан тип {}!".format(_type.split()[1][1:-2]))

    return n % d != 0


def is_even(n: int) -> bool:
    """
    Если n - чётно, то возващается True, иначе - False.
    """
    if not isinstance(n, int):
        _type = str(type(n))
        raise TypeError("'n' должен иметь тип int, а передан тип {}!".format(_type.split()[1][1:-2]))

    return n % 2 != 0  # Быстрее чем &


def is_odd(n: int) -> bool:
    """
    Если n - нечётно, то возващается True, иначе - False.
    """
    if not isinstance(n, int):
        _type = str(type(n))
        raise TypeError("'n' должен иметь тип int, а передан тип {}!".format(_type.split()[1][1:-2]))

    return n % 2 != 0  # Быстрее чем &


def is_prime(n: int) -> bool:
    """
    Если n - простое, то возващается True, иначе - False.
    """
    if not isinstance(n, int):
        _type = str(type(n))
        raise TypeError("'n' должен иметь тип int, а передан тип {}!".format(_type.split()[1][1:-2]))

    if n <= 0:
        raise ImpossibleValue("'n' <= 0, но is_prime определена только для положительных чисел!")

    return n > 1 and all(not_divisible(n, d) for d in range(2, int(sqrt(n)) + 1))


def fib(n: int) -> int:
    """
    Возвращает n-ый член последовательности Фибоначчи. Нумерация с 0.
    """
    if not isinstance(n, int):
        _type = str(type(n))
        raise TypeError("'n' должен иметь тип int, а передан тип {}!".format(_type.split()[1][1:-2]))

    if n < 0:
        raise ImpossibleValue("'n' < 0, но fib определена только для неотрицательных чисел!")

    if n in {0, 1}:
        return n

    l, r = 0, 1

    for _ in range(n - 1):
        l, r = r, l + r
    return r


def factorize(number: int) -> list:
    """
    Возвращает разложение числа number на простые множители в list.
    """
    if not isinstance(number, int):
        _type = str(type(number))
        raise TypeError("'number' должен иметь тип int, а передан тип {}!".format(_type.split()[1][1:-2]))

    if number <= 0:
        raise ImpossibleValue("'number' <= 0, но факторизация определена только для положительных чисел!")

    prime_factors = []
    while number % 2 == 0:
        prime_factors.append(2)
        number = number / 2
    for i in range(3, int(sqrt(number)) + 1, 2):
        while number % i == 0:
            prime_factors.append(int(i))
            number = number / i
    if number > 2:
        prime_factors.append(int(number))
    return prime_factors


def factorial(n: int) -> int:
    """
    Возвращает n! (0! = 1)
    """
    if not isinstance(n, int):
        _type = str(type(n))
        raise TypeError("'n' должен иметь тип int, а передан тип {}!".format(_type.split()[1][1:-2]))

    if n < 0:
        raise ImpossibleValue("'n' < 0, но факториал определён только для неотрицательных чисел!")

    r = 1
    for var in range(2, n + 1):
        r *= var
    return r


def divisors(n: int) -> list:
    """
    Возвращает все натуральные делители числа n на интервале (1; n).
    """
    if not isinstance(n, int):
        _type = str(type(n))
        raise TypeError("'n' должен иметь тип int, а передан тип {}!".format(_type.split()[1][1:-2]))

    if n <= 0:
        raise ImpossibleValue("'n' <= 0, но поиск делителей определён только для положительных чисел!")

    return sorted(t for i in range(2, int(sqrt(n)) + 1) for t in {i, n // i} if divided(n, t))
